import argparse
from gptools.stan import compile_model
from gptools.util import Timer
from gptools.util.fft import transform_irfft
from gptools.util.kernels import ExpQuadKernel
from gptools.util.graph import lattice_predecessors, predecessors_to_edge_index
from gptools.util.timeout import call_with_timeout
import numpy as np
import pathlib
import pickle
import tabulate
from tqdm import tqdm
from typing import List, Optional
from . import PARAMETERIZATIONS, sample_and_load_fit


def __main__(args: Optional[List[str]] = None) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("method", help="inference method to use", choices={"sample", "variational"})
    parser.add_argument("parameterization", help="parameterization of the model",
                        choices=PARAMETERIZATIONS)
    parser.add_argument("noise_scale", help="scale of observation noise", type=float)
    parser.add_argument("output", help="output path", nargs="?")
    parser.add_argument("--n", help="number of observations", type=int, default=100)
    parser.add_argument("--train_frac", help="fraction of points to use for training", type=float,
                        default=1.0)
    parser.add_argument("--num_parents", help="number of parents for GP on graphs", type=int,
                        default=5)
    parser.add_argument("--sigma", help="scale of Gaussian process covariance", type=float,
                        default=1.0)
    parser.add_argument("--length_scale", help="correlation length of Gaussian process covariance",
                        type=float, default=1.0)
    parser.add_argument("--epsilon", help="diagonal contribution to Gaussian process covariance",
                        type=float, default=1e-3)
    parser.add_argument("--seed", help="random number generator seed", type=int, default=42)
    parser.add_argument("--iter_sampling", help="number of posterior samples", type=int,
                        default=500)
    parser.add_argument("--show_progress", help="show progress bars", action="store_true")
    parser.add_argument("--show_diagnostics", help="show cmdstanpy diagnostics",
                        action="store_true")
    parser.add_argument("--iter_warmup", help="number of warmup samples", type=int)
    parser.add_argument("--timeout", help="timeout in seconds", type=float, default=60)
    parser.add_argument("--max_chains", type=int, default=1, help="maximum number of chains to "
                        "run; use -1 for an unlimited number of chains")
    parser.add_argument("--ignore_converged", action="store_true",
                        help="do not check if the variational algorithm has converged")
    args = parser.parse_args(args)

    # Compile the model.
    stan_file = pathlib.Path(__file__).parent / f"{args.parameterization}.stan"
    model = compile_model(stan_file=stan_file)

    # Prepare the results container.
    result = {"args": vars(args)}

    np.random.seed(args.seed)
    i = 0

    # We draw samples with Fourier methods because generating the training data can be the main
    # bottleneck if the sample size is large.
    kernel = ExpQuadKernel(args.sigma, args.length_scale, args.n)
    cov_rfft = kernel.evaluate_rfft(args.n) + args.epsilon

    with Timer() as total_timer, tqdm() as progress:
        while (args.max_chains == -1 or i < args.max_chains) \
                and (args.timeout is None or total_timer.duration < args.timeout):

            # Sample the Gaussian process.
            z = np.random.normal(0, 1, args.n)
            eta = transform_irfft(z, np.zeros_like(z), cov_rfft=cov_rfft)
            y = np.random.normal(eta, args.noise_scale)

            # Construct the nearest-neighbor graph.
            predecessors = lattice_predecessors((args.n,), args.num_parents)
            edge_index = predecessors_to_edge_index(predecessors)

            # Sample observed points and fit the model.
            num_observed = np.random.binomial(args.n, args.train_frac)
            observed_idx = np.random.choice(args.n, size=num_observed, replace=False) + 1
            data = {
                "n": args.n,
                "num_dims": 1,
                "X": np.arange(args.n)[:, None],
                "y": y,
                "sigma": args.sigma,
                "length_scale": args.length_scale,
                "epsilon": args.epsilon,
                "num_edges": edge_index.shape[1],
                "edge_index": edge_index,
                "noise_scale": args.noise_scale,
                "num_observed": num_observed,
                "observed_idx": observed_idx,
            }

            with Timer() as timer:
                try:
                    kwargs = {
                        "data": data,
                        "seed": args.seed,
                    }
                    if args.method == "sample":
                        iter_warmup = args.iter_warmup or args.iter_sampling
                        fit = call_with_timeout(
                            args.timeout, sample_and_load_fit, model,
                            iter_sampling=args.iter_sampling, chains=1, threads_per_chain=1,
                            show_progress=args.show_progress, iter_warmup=iter_warmup, **kwargs,
                        )
                    elif args.method == "variational":
                        fit = call_with_timeout(
                            args.timeout, model.variational, output_samples=args.iter_sampling,
                            require_converged=not args.ignore_converged, **kwargs,
                        )
                    else:  # pragma: no cover
                        raise ValueError(args.method)
                    timeout = False
                except TimeoutError:
                    fit = None
                    timeout = True

            result.setdefault("durations", []).append(timer.duration)
            result.setdefault("timeouts", []).append(timeout)
            result.setdefault("fits", []).append(fit)
            result.setdefault("data", []).append(data)
            result.setdefault("etas", []).append(eta)
            progress.update()
            i += 1

    for key in ["durations", "timeouts", "etas"]:
        result[key] = np.asarray(result[key])

    if args.output:
        with open(args.output, "wb") as fp:
            pickle.dump(result, fp)

    # Report the results.
    if all(result["timeouts"]):
        print(f"all chains timed out after {args.timeout:.3f} seconds")
        return

    # Show results on the first fit that didn't time out.
    for fit, timeout in zip(result["fits"], result["timeouts"]):
        if not timeout:
            break
    values = vars(args)
    values.update({
        "duration": f"{timer.duration:.3f}",
    })
    if args.method == "sample":
        values.update({
            "divergences": f"{fit.divergences.sum()} / {fit.num_draws_sampling} "
                f"({100 * fit.divergences.sum() / fit.num_draws_sampling:.1f}%)",  # noqa: E131
            "max_treedepths": f"{fit.max_treedepths.sum()} / {fit.num_draws_sampling} "
                f"({100 * fit.max_treedepths.sum() / fit.num_draws_sampling:.1f}%)",  # noqa: E131
        })
    rows = [(key, str(value)) for key, value in values.items()]
    print(tabulate.tabulate(rows))

    if args.method == "sample" and args.show_diagnostics:
        print(fit.diagnose())


if __name__ == "__main__":
    __main__()



import cmdstanpy
from cmdstanpy.model import OptionalPath
import logging
import os
from typing import Any, Dict, Optional, Type, Union


def get_include() -> str:
    """
    Get the include directory for the library.
    """
    return os.path.dirname(__file__)


def set_cmdstanpy_log_level(level: Optional[Union[str, int]] = None) -> None:
    """
    Set the log level of cmdstanpy.

    Args:
        level: Log level (defaults to `WARNING`).
    """
    cmdstanpy_logger = cmdstanpy.utils.get_logger()
    for handler in cmdstanpy_logger.handlers:
        handler.setLevel(level or logging.WARNING)


def compile_model(
        model_name: Optional[str] = None, stan_file: OptionalPath = None,
        exe_file: OptionalPath = None, compile: Union[bool, str] = True,
        stanc_options: Optional[Dict[str, Any]] = None,
        cpp_options: Optional[Dict[str, Any]] = None,
        user_header: OptionalPath = None, cls: Optional[Type] = None, **kwargs) \
            -> cmdstanpy.CmdStanModel:
    """
    Create a :class:`cmdstanpy.CmdStanModel` and configure include paths for gptools.

    Args:
        model_name: Model name used for output file names (defaults to base name of
            :code:`stan_file`).
        stan_file: Path to Stan model file.
        exe_file: Path to compiled executable file.  Optional unless :code:`stan_file` is not given.
            If both :code:`stan_file` and :code:`exe_file` are given, the base names must match.
        compile: Whether or not to compile the model (defaults to :code:`True`). If :code:`"force"`,
            the model will compile even if a newer executable is found.
        stanc_options: Stanc3 compiler options (see
            `here <https://mc-stan.org/docs/stan-users-guide/stanc-args.html>`__ for details).
        cpp_options: C++ compiler options.
        user_header: Path to a C++ header file to include during compilation.
        cls: Subclass of :class:`cmdstanpy.CmdStanModel` if a custom implementation should be used.
        **kwargs: Keyword arguments passed to the :class:`cmdstanpy.CmdStanModel` constructor.

    Returns:
        model: Model instance with include paths configured for gptools.
    """
    # Add gptools includes by default.
    stanc_options = stanc_options or {}
    stanc_options.setdefault("include-paths", []).append(get_include())
    cls = cls or cmdstanpy.CmdStanModel
    return cls(
        model_name=model_name, stan_file=stan_file, exe_file=exe_file, compile=compile,
        stanc_options=stanc_options, cpp_options=cpp_options, user_header=user_header, **kwargs,
    )


# Reduce logging verbosity if we are compiling documentation on readthedocs. We cannot easily
# achieve the same effect in the sphinx `conf.py`` because each example runs in its own kernel.
if "READTHEDOCS" in os.environ:  # pragma: no cover
    set_cmdstanpy_log_level()



from doit_interface import dict2args
from gptools.stan.profile import PARAMETERIZATIONS
from gptools.stan.profile.__main__ import __main__
import numpy as np
import pathlib
import pickle
import pytest
from typing import Tuple


def _run_main(tmp_path: pathlib.Path, method: str, parameterization: str, noise_scale: float = 1.0,
              **kwargs) -> Tuple[dict, dict]:
    """
    Run the main process.
    """
    defaults = {"iter_sampling": 3, "n": 7, "num_parents": 2}
    defaults.update(kwargs)
    kwargs = defaults
    path = tmp_path / "result.pkl"
    __main__([method, parameterization, str(noise_scale), str(path), "--show_diagnostics",
              "--ignore_converged", *dict2args(**kwargs)])
    with path.open("rb") as fp:
        return pickle.load(fp), kwargs


@pytest.mark.parametrize("method", ["sample", "variational"])
@pytest.mark.parametrize("parameterization", PARAMETERIZATIONS)
def test_profile(method: str, parameterization: str, tmp_path: pathlib.Path) -> None:
    result, kwargs = _run_main(tmp_path, method, parameterization)

    if method == "sample":
        variables = result["fits"][0].stan_variables()["eta"]
    elif method == "variational":
        variables = result["fits"][0].variational_sample.values[:, 3:]
        if "non_centered" in parameterization:
            _, variables = np.split(variables, 2, axis=1)
    else:
        raise ValueError(method)
    assert variables.shape == (kwargs["iter_sampling"], kwargs["n"])


def test_profile_timeout(tmp_path: pathlib.Path) -> None:
    result, kwargs = _run_main(tmp_path, "sample", "standard_centered", iter_sampling=100_000,
                               timeout=0.7)

    assert result["fits"][0] is None
    assert abs(result["durations"] - kwargs["timeout"]) < 0.1


def test_profile_max_chains(tmp_path: pathlib.Path) -> None:
    result, kwargs = _run_main(tmp_path, "sample", "standard_centered", max_chains=7)
    assert len(result["fits"]) == kwargs["max_chains"]



import re
from setuptools import find_namespace_packages, setup

with open("README.rst") as fp:
    long_description = fp.read()
long_description = re.sub(r".. (literalinclude|testsetup|toctree)::", "..\n    comment",
                          long_description)
long_description = re.sub(".. doctest::", ".. code-block::", long_description)
long_description = re.sub(":(doc|class|func|ref):", ":code:", long_description)

setup(
    name="gptools-stan",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=find_namespace_packages(),
    include_package_data=True,
    version="0.1.4",
    python_requires=">=3.8",
    install_requires=[
        # Required because of a bug in how complex numbers are handled (see
        # https://github.com/stan-dev/cmdstanpy/pull/612).
        "cmdstanpy>=1.0.7",
        "gptools-util",
        "numpy",
    ],
    extras_require={
        "docs": [
            "myst-nb",
            "networkx",
            "openpyxl",
            "pyproj",
            "sphinx",
            "sphinx-multiproject",
            "sphinx_rtd_theme",
            "sphinx-stan>=0.1.5",
        ],
        "tests": [
            "doit-interface",
            "flake8",
            "jupyter",
            "matplotlib",
            "pytest",
            "pytest-cov",
            "scipy",
            "tabulate",
            "twine",
        ],
    }
)



import cmdstanpy
from gptools.stan import set_cmdstanpy_log_level
import numpy as np


SIZES = 16 * 2 ** np.arange(11)
FOURIER_ONLY_SIZE_THRESHOLD = 16 * 2 ** 9
LOG10_NOISE_SCALES = np.linspace(-1, 1, 7)
PARAMETERIZATIONS = ["graph_centered", "graph_non_centered", "fourier_centered",
                     "fourier_non_centered", "standard_centered", "standard_non_centered"]

# Make cmdstanpy less verbose.
set_cmdstanpy_log_level()


def sample_and_load_fit(model: cmdstanpy.CmdStanModel, **kwargs) -> cmdstanpy.CmdStanMCMC:  # noqa: E501, pragma: no cover
    """
    Wrapper function to sample and load the data so the posterior samples can be serialized.
    """
    fit = model.sample(**kwargs)
    fit.stan_variables()
    return fit
